# -*- coding: utf-8 -*-
from .base import BaseMatcher
from requests.compat import urlparse


class HostMatcher(BaseMatcher):
    # Matches based on the host of the request
    name = 'host'

    def match(self, request, recorded_request):
        request_host = urlparse(request.url).netloc
        recorded_host = urlparse(recorded_request['uri']).netloc
        return request_host == recorded_host



# -*- coding: utf-8 -*-
"""A set of fixtures to integrate Betamax with py.test.

.. autofunction:: betamax_session

"""

from __future__ import absolute_import

import re
import warnings

import pytest
import requests

from .. import recorder as betamax


def _sanitize(name):
    """
    Replace certain characters (which might be problematic when contained in
    strings which will be used as file names) by '-'s. """
    return re.sub(r'[\s/<>:\\"|?*]', '-', name)


def _cassette_name(request, parametrized):
    """Determine a cassette name from request.

    :param request:
        A request object from pytest giving us context information for the
        fixture.
    :param parametrized:
        Whether the name should consider parametrized tests.
    :returns:
        A cassette name.
    """
    cassette_name = ''

    if request.module is not None:
        cassette_name += request.module.__name__ + '.'

    if request.cls is not None:
        cassette_name += request.cls.__name__ + '.'

    if parametrized:
        cassette_name += _sanitize(request.node.name)
    else:
        cassette_name += request.function.__name__
        if request.node.name != request.function.__name__:
            warnings.warn(
                "betamax_recorder and betamax_session currently don't include "
                "parameters in the cassette name. "
                "Use betamax_parametrized_recorder/_session to include "
                "parameters. "
                "This behavior will be the default in betamax 1.0",
                FutureWarning, stacklevel=3)

    return cassette_name


def _betamax_recorder(request, parametrized=True):
    cassette_name = _cassette_name(request, parametrized=parametrized)
    session = requests.Session()
    recorder = betamax.Betamax(session)
    recorder.use_cassette(cassette_name)
    recorder.start()
    request.addfinalizer(recorder.stop)
    return recorder


@pytest.fixture
def betamax_recorder(request):
    """Generate a recorder with a session that has Betamax already installed.

    This will create a new Betamax instance with a generated cassette name.
    The cassette name is generated by first using the module name from where
    the test is collected, then the class name (if it exists), and then the
    test function name. For example, if your test is in ``test_stuff.py`` and
    is the method ``TestStuffClass.test_stuff`` then your cassette name will be
    ``test_stuff_TestStuffClass_test_stuff``. If the test is parametrized,
    the parameters will not be included in the name. In case you need that,
    use betamax_parametrized_recorder instead. This will change in 1.0.0,
    where parameters will be included by default.

    :param request:
        A request object from pytest giving us context information for the
        fixture.
    :returns:
        An instantiated recorder.
    """
    return _betamax_recorder(request, parametrized=False)


@pytest.fixture
def betamax_session(betamax_recorder):
    """Generate a session that has Betamax already installed.

    See `betamax_recorder` fixture.

    :param betamax_recorder:
        A recorder fixture with a configured request session.
    :returns:
        An instantiated requests Session wrapped by Betamax.
    """

    return betamax_recorder.session


@pytest.fixture
def betamax_parametrized_recorder(request):
    """Generate a recorder with a session that has Betamax already installed.

    This will create a new Betamax instance with a generated cassette name.
    The cassette name is generated by first using the module name from where
    the test is collected, then the class name (if it exists), and then the
    test function name with parameters if parametrized.
    For example, if your test is in ``test_stuff.py`` and
    the method is ``TestStuffClass.test_stuff`` with parameter ``True`` then
    your cassette name will be
    ``test_stuff_TestStuffClass_test_stuff[True]``.

    :param request:
        A request object from pytest giving us context information for the
        fixture.
    :returns:
        An instantiated recorder.
    """
    warnings.warn(
        "betamax_parametrized_recorder and betamax_parametrized_session "
        "will be removed in betamax 1.0. Their behavior will be the "
        "default.",
        DeprecationWarning)
    return _betamax_recorder(request, parametrized=True)


@pytest.fixture
def betamax_parametrized_session(betamax_parametrized_recorder):
    """Generate a session that has Betamax already installed.

    See `betamax_parametrized_recorder` fixture.

    :param betamax_parametrized_recorder:
        A recorder fixture with a configured request session.
    :returns:
        An instantiated requests Session wrapped by Betamax.
    """

    return betamax_parametrized_recorder.session



from requests.cookies import extract_cookies_to_jar
from datetime import datetime

from betamax import util


class Interaction(object):

    """The Interaction object represents the entirety of a single interaction.

    The interaction includes the date it was recorded, its JSON
    representation, and the ``requests.Response`` object complete with its
    ``request`` attribute.

    This object also handles the filtering of sensitive data.

    No methods or attributes on this object are considered public or part of
    the public API. As such they are entirely considered implementation
    details and subject to change. Using or relying on them is not wise or
    advised.

    """

    def __init__(self, interaction, response=None):
        self.data = interaction
        self.orig_response = response
        self.recorded_response = self.deserialize()
        self.used = False
        self.ignored = False

    def ignore(self):
        """Ignore this interaction.

        This is only to be used from a before_record or a before_playback
        callback.
        """
        self.ignored = True

    def as_response(self):
        """Return the Interaction as a Response object."""
        self.recorded_response = self.deserialize()
        return self.recorded_response

    @property
    def recorded_at(self):
        return datetime.strptime(self.data['recorded_at'], '%Y-%m-%dT%H:%M:%S')

    def deserialize(self):
        """Turn a serialized interaction into a Response."""
        r = util.deserialize_response(self.data['response'])
        r.request = util.deserialize_prepared_request(self.data['request'])
        extract_cookies_to_jar(r.cookies, r.request, r.raw)
        return r

    def match(self, matchers):
        """Return whether this interaction is a match."""
        request = self.data['request']
        return all(m(request) for m in matchers)

    def replace(self, text_to_replace, placeholder):
        """Replace sensitive data in this interaction."""
        self.replace_in_headers(text_to_replace, placeholder)
        self.replace_in_body(text_to_replace, placeholder)
        self.replace_in_uri(text_to_replace, placeholder)

    def replace_all(self, replacements, serializing):
        """Easy way to accept all placeholders registered."""
        for placeholder in replacements:
            self.replace(*placeholder.unpack(serializing))

    def replace_in_headers(self, text_to_replace, placeholder):
        if text_to_replace == '':
            return
        for obj in ('request', 'response'):
            headers = self.data[obj]['headers']
            for k, v in list(headers.items()):
                if isinstance(v, list):
                    headers[k] = [hv.replace(text_to_replace, placeholder)
                                  for hv in v]
                else:
                    headers[k] = v.replace(text_to_replace, placeholder)

    def replace_in_body(self, text_to_replace, placeholder):
        if text_to_replace == '':
            return
        for obj in ('request', 'response'):
            body = self.data[obj]['body']
            old_style = hasattr(body, 'replace')
            if not old_style:
                body = body.get('string', '')

            if text_to_replace in body:
                body = body.replace(text_to_replace, placeholder)
            if old_style:
                self.data[obj]['body'] = body
            else:
                self.data[obj]['body']['string'] = body

    def replace_in_uri(self, text_to_replace, placeholder):
        if text_to_replace == '':
            return
        for (obj, key) in (('request', 'uri'), ('response', 'url')):
            uri = self.data[obj][key]
            if text_to_replace in uri:
                self.data[obj][key] = uri.replace(
                    text_to_replace, placeholder
                )



# -*- coding: utf-8 -*-
import collections
from datetime import datetime
from functools import partial
import os.path

from .interaction import Interaction

from .. import matchers
from .. import serializers
from betamax.util import (_option_from, serialize_prepared_request,
                          serialize_response, timestamp)


class Cassette(object):

    default_cassette_options = {
        'record_mode': 'once',
        'match_requests_on': ['method', 'uri'],
        're_record_interval': None,
        'placeholders': [],
        'preserve_exact_body_bytes': False,
        'allow_playback_repeats': False,
    }

    hooks = collections.defaultdict(list)

    def __init__(self, cassette_name, serialization_format, **kwargs):
        #: Short name of the cassette
        self.cassette_name = cassette_name

        self.serialized = None

        defaults = Cassette.default_cassette_options

        # Determine the record mode
        self.record_mode = _option_from('record_mode', kwargs, defaults)

        # Retrieve the serializer for this cassette
        self.serializer = serializers.SerializerProxy.find(
            serialization_format, kwargs.get('cassette_library_dir'),
            cassette_name
            )
        self.cassette_path = self.serializer.cassette_path

        # Determine which placeholders to use
        default_placeholders = defaults['placeholders'][:]
        cassette_placeholders = kwargs.get('placeholders', [])
        self.placeholders = merge_placeholder_lists(default_placeholders,
                                                    cassette_placeholders)

        # Determine whether to preserve exact body bytes
        self.preserve_exact_body_bytes = _option_from(
            'preserve_exact_body_bytes', kwargs, defaults
            )

        self.allow_playback_repeats = _option_from(
            'allow_playback_repeats', kwargs, defaults
            )

        # Initialize the interactions
        self.interactions = []

        # Initialize the match options
        self.match_options = set()

        self.load_interactions()
        self.serializer.allow_serialization = self.is_recording()

    @staticmethod
    def can_be_loaded(cassette_library_dir, cassette_name, serialize_with,
                      record_mode):
        # If we want to record a cassette we don't care if the file exists
        # yet
        recording = False
        if record_mode in ['once', 'all', 'new_episodes']:
            recording = True

        serializer = serializers.serializer_registry.get(
            serialize_with
        )
        if not serializer:
            raise ValueError(
                'Serializer {0} is not registered with Betamax'.format(
                    serialize_with
                    ))

        cassette_path = serializer.generate_cassette_name(
            cassette_library_dir, cassette_name
            )
        # Otherwise if we're only replaying responses, we should probably
        # have the cassette the user expects us to load and raise.
        return os.path.exists(cassette_path) or recording

    def clear(self):
        # Clear out the interactions
        self.interactions = []
        # Serialize to the cassette file
        self._save_cassette()

    @property
    def earliest_recorded_date(self):
        """The earliest date of all of the interactions this cassette."""
        if self.interactions:
            i = sorted(self.interactions, key=lambda i: i.recorded_at)[0]
            return i.recorded_at
        return datetime.now()

    def eject(self):
        self._save_cassette()

    def find_match(self, request):
        """Find a matching interaction based on the matchers and request.

        This uses all of the matchers selected via configuration or
        ``use_cassette`` and passes in the request currently in progress.

        :param request: ``requests.PreparedRequest``
        :returns: :class:`~betamax.cassette.Interaction`
        """
        # if we are recording, do not filter by match
        if self.is_recording():
            if ((self.record_mode == 'new_episodes' and
                 all(i.used is True for i in self.interactions)) or
                    self.record_mode in ('once', 'none')):
                return None

        opts = self.match_options
        # Curry those matchers
        curried_matchers = [
            partial(matchers.matcher_registry[o].match, request)
            for o in opts
        ]

        for interaction in self.interactions:
            if not interaction.match(curried_matchers):
                continue

            if interaction.used or interaction.ignored:
                continue

            # If the interaction matches everything
            if self.record_mode == 'all':
                # If we're recording everything and there's a matching
                # interaction we want to overwrite it, so we remove it.
                self.interactions.remove(interaction)
                break

            # set interaction as used before returning
            if not self.allow_playback_repeats:
                interaction.used = True
            return interaction

        # No matches. So sad.
        return None

    def is_empty(self):
        """Determine if the cassette was empty when loaded."""
        return not self.serialized

    def is_recording(self):
        """Return whether the cassette is recording."""
        values = {
            'none': False,
            'once': self.is_empty(),
        }
        return values.get(self.record_mode, True)

    def load_interactions(self):
        if self.serialized is None:
            self.serialized = self.serializer.deserialize()

        interactions = self.serialized.get('http_interactions', [])
        self.interactions = [Interaction(i) for i in interactions]

        for i in self.interactions:
            dispatch_hooks('before_playback', i, self)
            i.replace_all(self.placeholders, False)

    def sanitize_interactions(self):
        for i in self.interactions:
            i.replace_all(self.placeholders, True)

    def save_interaction(self, response, request):
        serialized_data = self.serialize_interaction(response, request)
        interaction = Interaction(serialized_data, response)
        dispatch_hooks('before_record', interaction, self)
        if not interaction.ignored:  # If a hook caused this to be ignored
            self.interactions.append(interaction)
        return interaction

    def serialize_interaction(self, response, request):
        return {
            'request': serialize_prepared_request(
                request,
                self.preserve_exact_body_bytes
                ),
            'response': serialize_response(
                response,
                self.preserve_exact_body_bytes
                ),
            'recorded_at': timestamp(),
        }

    # Private methods
    def _save_cassette(self):
        from .. import __version__
        self.sanitize_interactions()

        cassette_data = {
            'http_interactions': [i.data for i in self.interactions],
            'recorded_with': 'betamax/{0}'.format(__version__)
        }
        self.serializer.serialize(cassette_data)


class Placeholder(collections.namedtuple('Placeholder',
                                         'placeholder replace')):
    """Encapsulate some logic about Placeholders."""

    @classmethod
    def from_dict(cls, dictionary):
        return cls(**dictionary)

    def unpack(self, serializing):
        if serializing:
            return self.replace, self.placeholder
        else:
            return self.placeholder, self.replace


def merge_placeholder_lists(defaults, overrides):
    overrides = [Placeholder.from_dict(override) for override in overrides]
    overrides_dict = dict((p.placeholder, p) for p in overrides)
    placeholders = [overrides_dict.pop(p.placeholder, p)
                    for p in map(Placeholder.from_dict, defaults)]
    return placeholders + [p for p in overrides
                           if p.placeholder in overrides_dict]


def dispatch_hooks(hook_name, *args):
    """Dispatch registered hooks."""
    # Cassette.hooks is a dictionary that defaults to an empty list,
    # we neither need to check for the presence of hook_name in it, nor
    # need to worry about whether the return value will be iterable
    hooks = Cassette.hooks[hook_name]
    for hook in hooks:
        hook(*args)



"""Backport of urllib3's HTTPHeaderDict for older versions of requests.

This code was originally licensed under the MIT License and is copyrighted by
Andrey Petrov and contributors to urllib3.

This version was imported from:
https://github.com/shazow/urllib3/blob/3bd63406bef7c16d007c17563b6af14582567d4b/urllib3/_collections.py
"""
import sys

from collections import Mapping, MutableMapping

__all__ = ('HTTPHeaderDict',)

PY3 = sys.version_info >= (3, 0)


class HTTPHeaderDict(MutableMapping):
    """
    :param headers:
        An iterable of field-value pairs. Must not contain multiple field names
        when compared case-insensitively.

    :param kwargs:
        Additional field-value pairs to pass in to ``dict.update``.

    A ``dict`` like container for storing HTTP Headers.

    Field names are stored and compared case-insensitively in compliance with
    RFC 7230. Iteration provides the first case-sensitive key seen for each
    case-insensitive pair.

    Using ``__setitem__`` syntax overwrites fields that compare equal
    case-insensitively in order to maintain ``dict``'s api. For fields that
    compare equal, instead create a new ``HTTPHeaderDict`` and use ``.add``
    in a loop.

    If multiple fields that are equal case-insensitively are passed to the
    constructor or ``.update``, the behavior is undefined and some will be
    lost.

    >>> headers = HTTPHeaderDict()
    >>> headers.add('Set-Cookie', 'foo=bar')
    >>> headers.add('set-cookie', 'baz=quxx')
    >>> headers['content-length'] = '7'
    >>> headers['SET-cookie']
    'foo=bar, baz=quxx'
    >>> headers['Content-Length']
    '7'
    """

    def __init__(self, headers=None, **kwargs):
        super(HTTPHeaderDict, self).__init__()
        self._container = {}
        if headers is not None:
            if isinstance(headers, HTTPHeaderDict):
                self._copy_from(headers)
            else:
                self.extend(headers)
        if kwargs:
            self.extend(kwargs)

    def __setitem__(self, key, val):
        self._container[key.lower()] = (key, val)
        return self._container[key.lower()]

    def __getitem__(self, key):
        val = self._container[key.lower()]
        return ', '.join(val[1:])

    def __delitem__(self, key):
        del self._container[key.lower()]

    def __contains__(self, key):
        return key.lower() in self._container

    def __eq__(self, other):
        if not isinstance(other, Mapping) and not hasattr(other, 'keys'):
            return False
        if not isinstance(other, type(self)):
            other = type(self)(other)
        return (dict((k.lower(), v) for k, v in self.itermerged()) ==
                dict((k.lower(), v) for k, v in other.itermerged()))

    def __ne__(self, other):
        return not self.__eq__(other)

    if not PY3:  # Python 2
        iterkeys = MutableMapping.iterkeys
        itervalues = MutableMapping.itervalues

    __marker = object()

    def __len__(self):
        return len(self._container)

    def __iter__(self):
        # Only provide the originally cased names
        for vals in self._container.values():
            yield vals[0]

    def pop(self, key, default=__marker):
        """D.pop(k[,d]) -> v, remove specified key and return the value.

        If key is not found, d is returned if given, otherwise KeyError is
        raised.
        """
        # Using the MutableMapping function directly fails due to the private
        # marker.
        # Using ordinary dict.pop would expose the internal structures.
        # So let's reinvent the wheel.
        try:
            value = self[key]
        except KeyError:
            if default is self.__marker:
                raise
            return default
        else:
            del self[key]
            return value

    def discard(self, key):
        try:
            del self[key]
        except KeyError:
            pass

    def add(self, key, val):
        """Adds a (name, value) pair, doesn't overwrite the value if it already
        exists.

        >>> headers = HTTPHeaderDict(foo='bar')
        >>> headers.add('Foo', 'baz')
        >>> headers['foo']
        'bar, baz'
        """
        key_lower = key.lower()
        new_vals = key, val
        # Keep the common case aka no item present as fast as possible
        vals = self._container.setdefault(key_lower, new_vals)
        if new_vals is not vals:
            # new_vals was not inserted, as there was a previous one
            if isinstance(vals, list):
                # If already several items got inserted, we have a list
                vals.append(val)
            else:
                # vals should be a tuple then, i.e. only one item so far
                # Need to convert the tuple to list for further extension
                self._container[key_lower] = [vals[0], vals[1], val]

    def extend(self, *args, **kwargs):
        """Generic import function for any type of header-like object.
        Adapted version of MutableMapping.update in order to insert items
        with self.add instead of self.__setitem__
        """
        if len(args) > 1:
            raise TypeError("extend() takes at most 1 positional "
                            "arguments ({0} given)".format(len(args)))
        other = args[0] if len(args) >= 1 else ()

        if isinstance(other, HTTPHeaderDict):
            for key, val in other.iteritems():
                self.add(key, val)
        elif isinstance(other, Mapping):
            for key in other:
                self.add(key, other[key])
        elif hasattr(other, "keys"):
            for key in other.keys():
                self.add(key, other[key])
        else:
            for key, value in other:
                self.add(key, value)

        for key, value in kwargs.items():
            self.add(key, value)

    def getlist(self, key):
        """Returns a list of all the values for the named field. Returns an
        empty list if the key doesn't exist."""
        try:
            vals = self._container[key.lower()]
        except KeyError:
            return []
        else:
            if isinstance(vals, tuple):
                return [vals[1]]
            else:
                return vals[1:]

    # Backwards compatibility for httplib
    getheaders = getlist
    getallmatchingheaders = getlist
    iget = getlist

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, dict(self.itermerged()))

    def _copy_from(self, other):
        for key in other:
            val = other.getlist(key)
            if isinstance(val, list):
                # Don't need to convert tuples
                val = list(val)
            self._container[key.lower()] = [key] + val

    def copy(self):
        clone = type(self)()
        clone._copy_from(self)
        return clone

    def iteritems(self):
        """Iterate over all header lines, including duplicate ones."""
        for key in self:
            vals = self._container[key.lower()]
            for val in vals[1:]:
                yield vals[0], val

    def itermerged(self):
        """Iterate over all headers, merging duplicate ones together."""
        for key in self:
            val = self._container[key.lower()]
            yield val[0], ', '.join(val[1:])

    def items(self):
        return list(self.iteritems())

    @classmethod
    def from_httplib(cls, message):  # Python 2
        """Read headers from a Python 2 httplib message object."""
        # python2.7 does not expose a proper API for exporting multiheaders
        # efficiently. This function re-reads raw lines from the message
        # object and extracts the multiheaders properly.
        headers = []

        for line in message.headers:
            if line.startswith((' ', '\t')):
                key, value = headers[-1]
                headers[-1] = (key, value + '\r\n' + line.rstrip())
                continue

            key, value = line.split(':', 1)
            headers.append((key, value.strip()))

        return cls(headers)



"""
betamax.

=======

See https://betamax.readthedocs.io/ for documentation.

:copyright: (c) 2013-2018 by Ian Stapleton Cordasco
:license: Apache 2.0, see LICENSE for more details

"""

from .decorator import use_cassette
from .exceptions import BetamaxError
from .matchers import BaseMatcher
from .recorder import Betamax
from .serializers import BaseSerializer

__all__ = ('BetamaxError', 'Betamax', 'BaseMatcher', 'BaseSerializer',
           'use_cassette')
__author__ = 'Ian Stapleton Cordasco'
__copyright__ = 'Copyright 2013- Ian Stapleton Cordasco'
__license__ = 'Apache 2.0'
__title__ = 'betamax'
__version__ = '0.9.0'
__version_info__ = tuple(int(i) for i in __version__.split('.'))



class BetamaxError(Exception):
    def __init__(self, message):
        super(BetamaxError, self).__init__(message)


class MissingDirectoryError(BetamaxError):
    pass


class ValidationError(BetamaxError):
    pass


class InvalidOption(ValidationError):
    pass


class BodyBytesValidationError(ValidationError):
    pass


class MatchersValidationError(ValidationError):
    pass


class RecordValidationError(ValidationError):
    pass


class RecordIntervalValidationError(ValidationError):
    pass


class PlaceholdersValidationError(ValidationError):
    pass


class PlaybackRepeatsValidationError(ValidationError):
    pass


class SerializerValidationError(ValidationError):
    pass


validation_error_map = {
    'allow_playback_repeats': PlaybackRepeatsValidationError,
    'match_requests_on': MatchersValidationError,
    'record': RecordValidationError,
    'placeholders': PlaceholdersValidationError,
    'preserve_exact_body_bytes': BodyBytesValidationError,
    're_record_interval': RecordIntervalValidationError,
    'serialize': SerializerValidationError,  # TODO: Remove this
    'serialize_with': SerializerValidationError
}



# -*- coding: utf-8 -*-
from .base import BaseMatcher


class MethodMatcher(BaseMatcher):
    # Matches based on the method of the request
    name = 'method'

    def match(self, request, recorded_request):
        return request.method == recorded_request['method']



# -*- coding: utf-8 -*-
from .base import BaseMatcher
from requests.compat import urlparse


class PathMatcher(BaseMatcher):
    # Matches based on the path of the request
    name = 'path'

    def match(self, request, recorded_request):
        request_path = urlparse(request.url).path
        recorded_path = urlparse(recorded_request['uri']).path
        return request_path == recorded_path



from .base import BaseMatcher
from .body import BodyMatcher
from .digest_auth import DigestAuthMatcher
from .headers import HeadersMatcher
from .host import HostMatcher
from .method import MethodMatcher
from .path import PathMatcher
from .query import QueryMatcher
from .uri import URIMatcher

matcher_registry = {}

__all__ = ('BaseMatcher', 'BodyMatcher', 'DigestAuthMatcher',
           'HeadersMatcher', 'HostMatcher', 'MethodMatcher', 'PathMatcher',
           'QueryMatcher', 'URIMatcher', 'matcher_registry')


_matchers = [BodyMatcher, DigestAuthMatcher, HeadersMatcher, HostMatcher,
             MethodMatcher, PathMatcher, QueryMatcher, URIMatcher]
matcher_registry.update(dict((m.name, m()) for m in _matchers))
del _matchers



from email import parser, message
import sys


class MockHTTPResponse(object):
    def __init__(self, headers):
        from betamax.util import coerce_content

        h = ["%s: %s" % (k, v) for k in headers for v in headers.getlist(k)]
        h = map(coerce_content, h)
        h = '\r\n'.join(h)
        if sys.version_info < (2, 7):
            h = h.encode()
        p = parser.Parser(EmailMessage)
        # Thanks to Python 3, we have to use the slightly more awful API below
        # mimetools was deprecated so we have to use email.message.Message
        # which takes no arguments in its initializer.
        self.msg = p.parsestr(h)
        self.msg.set_payload(h)

        self._closed = False

    def isclosed(self):
        return self._closed

    def close(self):
        self._closed = True


class EmailMessage(message.Message):
    def getheaders(self, value, *args):
        return self.get_all(value, [])



from .cassette import Cassette
from .exceptions import InvalidOption, validation_error_map


def validate_record(record):
    return record in ['all', 'new_episodes', 'none', 'once']


def validate_matchers(matchers):
    from betamax.matchers import matcher_registry
    available_matchers = list(matcher_registry.keys())
    return all(m in available_matchers for m in matchers)


def validate_serializer(serializer):
    from betamax.serializers import serializer_registry
    return serializer in list(serializer_registry.keys())


def validate_placeholders(placeholders):
    """Validate placeholders is a dict-like structure"""
    keys = ['placeholder', 'replace']
    try:
        return all(sorted(list(p.keys())) == keys for p in placeholders)
    except TypeError:
        return False


def translate_cassette_options():
    for (k, v) in Cassette.default_cassette_options.items():
        yield (k, v) if k != 'record_mode' else ('record', v)


def isboolean(value):
    return value in [True, False]


class Options(object):
    valid_options = {
        'match_requests_on': validate_matchers,
        're_record_interval': lambda x: x is None or x > 0,
        'record': validate_record,
        'serialize': validate_serializer,  # TODO: Remove this
        'serialize_with': validate_serializer,
        'preserve_exact_body_bytes': isboolean,
        'placeholders': validate_placeholders,
        'allow_playback_repeats': isboolean,
    }

    defaults = {
        'match_requests_on': ['method', 'uri'],
        're_record_interval': None,
        'record': 'once',
        'serialize': None,  # TODO: Remove this
        'serialize_with': 'json',
        'preserve_exact_body_bytes': False,
        'placeholders': [],
        'allow_playback_repeats': False,
    }

    def __init__(self, data=None):
        self.data = data or {}
        self.validate()
        self.defaults = Options.defaults.copy()
        self.defaults.update(translate_cassette_options())

    def __repr__(self):
        return 'Options(%s)' % self.data

    def __getitem__(self, key):
        return self.data.get(key, self.defaults.get(key))

    def __setitem__(self, key, value):
        self.data[key] = value
        return value

    def __delitem__(self, key):
        del self.data[key]

    def __contains__(self, key):
        return key in self.data

    def items(self):
        return self.data.items()

    def validate(self):
        for key, value in list(self.data.items()):
            if key not in Options.valid_options:
                raise InvalidOption('{0} is not a valid option'.format(key))
            else:
                is_valid = Options.valid_options[key]
                if not is_valid(value):
                    raise validation_error_map[key]('{0!r} is not valid'
                                                    .format(value))



# -*- coding: utf-8 -*-
from .base import BaseSerializer
from .json_serializer import JSONSerializer
from .proxy import SerializerProxy

serializer_registry = {}

_serializers = [JSONSerializer]
serializer_registry.update(dict((s.name, s()) for s in _serializers))
del _serializers

__all__ = ('BaseSerializer', 'JSONSerializer', 'SerializerProxy')



# -*- coding: utf-8 -*-
from .base import BaseMatcher

from betamax import util


class BodyMatcher(BaseMatcher):
    # Matches based on the body of the request
    name = 'body'

    def match(self, request, recorded_request):
        recorded_request = util.deserialize_prepared_request(recorded_request)

        request_body = b''
        if request.body:
            request_body = util.coerce_content(request.body)

        recorded_body = b''
        if recorded_request.body:
            recorded_body = util.coerce_content(recorded_request.body)

        return recorded_body == request_body



"""Minimal :class:`unittest.TestCase` subclass adding Betamax integration.

.. autoclass:: betamax.fixtures.unittest.BetamaxTestCase
    :members:

When using Betamax with unittest, you can use the traditional style of Betamax
covered in the documentation thoroughly, or you can use your fixture methods,
:meth:`unittest.TestCase.setUp` and :meth:`unittest.TestCase.tearDown` to wrap
entire tests in Betamax.

Here's how you might use it:

.. code-block:: python

    from betamax.fixtures import unittest

    from myapi import SessionManager


    class TestMyApi(unittest.BetamaxTestCase):
        def setUp(self):
            # Call BetamaxTestCase's setUp first to get a session
            super(TestMyApi, self).setUp()

            self.manager = SessionManager(self.session)

        def test_all_users(self):
            \"\"\"Retrieve all users from the API.\"\"\"
            for user in self.manager:
                # Make assertions or something

Alternatively, if you are subclassing a :class:`requests.Session` to provide
extra functionality, you can do something like this:

.. code-block:: python

    from betamax.fixtures import unittest

    from myapi import Session, SessionManager


    class TestMyApi(unittest.BetamaxTestCase):
        SESSION_CLASS = Session

        # See above ...

"""
# NOTE(sigmavirus24): absolute_import is required to make import unittest work
from __future__ import absolute_import

try:
    import unittest2 as unittest
except ImportError:
    import unittest

import requests

from .. import recorder


__all__ = ('BetamaxTestCase',)


class BetamaxTestCase(unittest.TestCase):

    """Betamax integration for unittest.

    .. versionadded:: 0.5.0
    """

    #: Class that is a subclass of :class:`requests.Session`
    SESSION_CLASS = requests.Session

    #: Custom path to save cassette.
    CASSETTE_LIBRARY_DIR = None

    def generate_cassette_name(self):
        """Generates a cassette name for the current test.

        The default format is "%(classname)s.%(testMethodName)s"

        To change the default cassette format, override this method in a
        subclass.

        :returns: Cassette name for the current test.
        :rtype: str
        """
        cls = getattr(self, '__class__')
        test = self._testMethodName
        return '{0}.{1}'.format(cls.__name__, test)

    def setUp(self):
        """Betamax-ified setUp fixture.

        This will call the superclass' setUp method *first* and then it will
        create a new :class:`requests.Session` and wrap that in a Betamax
        object to record it. At the end of ``setUp``, it will start recording.
        """
        # Bail out early if the SESSION_CLASS isn't a subclass of
        # requests.Session
        self.assertTrue(issubclass(self.SESSION_CLASS, requests.Session))
        # Make sure if the user is multiply inheriting that all setUps are
        # called. (If that confuses you, see: https://youtu.be/EiOglTERPEo)
        super(BetamaxTestCase, self).setUp()

        cassette_name = self.generate_cassette_name()

        self.session = self.SESSION_CLASS()
        self.recorder = recorder.Betamax(
            session=self.session,
            cassette_library_dir=self.CASSETTE_LIBRARY_DIR)

        self.recorder.use_cassette(cassette_name)
        self.recorder.start()

    def tearDown(self):
        """Betamax-ified tearDown fixture.

        This will call the superclass' tearDown method *first* and then it
        will stop recording interactions.
        """
        super(BetamaxTestCase, self).tearDown()
        self.recorder.stop()
